from autofit.core import model_mapper as mm
from autocti.model import arctic_params as ap
import workspace_jam.scripts.requirements as req
import workspace_jam.scripts.requirement_converter as convert

import numpy as np

model_mapper = mm.ModelMapper(serial=ap.SerialThreeSpecies)
model_mapper.serial.well_fill_alpha = 0.0
model_mapper.serial.well_fill_gamma = 0.0

input_model = [0.01, 0.03, 0.9, 0.8, 3.5, 20.0, 1.0e-4, 0.58]
input_rho = np.array([0.01, 0.03, 0.9])
input_tau = np.array([0.8, 3.5, 20.0])

sigma_limit = 2.0

most_likely, most_probable, lower_limits, upper_limits, lower_error, upper_error, error_pecision = \
    req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model, sigma_limit=sigma_limit,
                                name='Results/Serial_Fix_lr/')

delta_ellipticity_true = convert.convert_to_ellipticity(rho=input_rho, tau=input_tau)

model_ml_rho = np.array([most_likely[2], most_likely[3], most_likely[4]])
model_ml_tau = np.array([most_likely[5], most_likely[6], most_likely[7]])
delta_ellipticity_model_ml = convert.convert_to_ellipticity(rho=model_ml_rho, tau=model_ml_tau)

model_rho = np.array([most_probable[0], most_probable[1], most_probable[2]])
model_tau = np.array([most_probable[3], most_probable[4], most_probable[5]])
delta_ellipticity_model = convert.convert_to_ellipticity(rho=model_rho, tau=model_tau)

upper_rho = np.array([upper_limits[0], upper_limits[1], upper_limits[2]])
upper_tau = np.array([upper_limits[3], upper_limits[4], upper_limits[5]])
delta_ellipticity_upper = convert.convert_to_ellipticity(rho=upper_rho, tau=upper_tau)

lower_rho = np.array([lower_limits[0], lower_limits[1], lower_limits[2]])
lower_tau = np.array([lower_limits[3], lower_limits[4], lower_limits[5]])
delta_ellipticity_lower = convert.convert_to_ellipticity(rho=lower_rho, tau=lower_tau)

print('Model True')
print(delta_ellipticity_true)
print(np.sum(delta_ellipticity_true))
print()

print('Model most probable')
print(delta_ellipticity_model)
print(np.sum(delta_ellipticity_model))
print()

print('Model upper')
print(delta_ellipticity_upper)
print(np.sum(delta_ellipticity_upper))
print()

print('Model lower')
print(delta_ellipticity_lower)
print(np.sum(delta_ellipticity_lower))
print()

print('Induced Delta Elliptcitiy')
print(np.abs(np.sum(delta_ellipticity_model) - np.sum(delta_ellipticity_true)))
print(np.abs(np.sum(delta_ellipticity_upper) - np.sum(delta_ellipticity_true)))
print(np.abs(np.sum(delta_ellipticity_lower) - np.sum(delta_ellipticity_true)))
print()

print('Induced Delta Elliptcitiy / Requirement')
print(np.abs(np.sum(delta_ellipticity_model) - np.sum(delta_ellipticity_true)) / (1.1*10e-4))
print(np.abs(np.sum(delta_ellipticity_upper) - np.sum(delta_ellipticity_true)) / (1.1*10e-4))
print(np.abs(np.sum(delta_ellipticity_lower) - np.sum(delta_ellipticity_true)) / (1.1*10e-4))


import numpy as np

from autocti.charge_injection import ci_data
from autocti.charge_injection import ci_fit
from autocti.charge_injection.plotters import ci_fit_plotters
from autocti.data import mask as msk
from autocti.model import arctic_params
from autocti.model import arctic_settings
from test.charge_injection.plotters.fixtures import *
from test.mock.mock import MockPattern, MockCIFrame


@pytest.fixture(name='ci_fit_plotter_path')
def make_ci_data_plotter_setup():
    return "{}/../../test_files/plotting/ci_fit/".format(os.path.dirname(os.path.realpath(__file__)))


@pytest.fixture(name='mask')
def make_mask():
    return msk.Mask.empty_for_shape(shape=(2, 2))


@pytest.fixture(name='image')
def make_image():
    return np.ones((2, 2))


@pytest.fixture(name='noise_map')
def make_noise_map():
    return 2.0 * np.ones((2, 2))


@pytest.fixture(name='ci_pre_cti')
def make_ci_pre_cti():
    return 3.0 * np.ones((2, 2))


@pytest.fixture(name='ci_data_fit')
def make_ci_data_fit(image, noise_map, mask, ci_pre_cti):
    return ci_data.MaskedCIData(image=image, noise_map=noise_map, ci_pre_cti=ci_pre_cti, mask=mask,
                                ci_pattern=MockPattern(), ci_frame=MockCIFrame(value=3.0))


@pytest.fixture(name="cti_settings")
def make_cti_settings():
    parallel_settings = arctic_settings.Settings(well_depth=0, niter=1, express=1, n_levels=2000)
    return arctic_settings.ArcticSettings(neomode='NEO', parallel=parallel_settings)


@pytest.fixture(name="cti_params")
def make_cti_params():
    parallel_1_species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
    return arctic_params.ArcticParams(parallel_species=parallel_1_species)


@pytest.fixture(name="fit")
def make_fit(ci_data_fit, cti_params, cti_settings):
    return ci_fit.CIFit(masked_ci_data=ci_data_fit, cti_params=cti_params,
                                                               cti_settings=cti_settings)


def test__fit_individuals__depedent_on_input(fit, ci_fit_plotter_path, plot_patch):
    ci_fit_plotters.plot_fit_individuals(
        fit=fit,
        should_plot_image=True,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_ci_pre_cti=True,
        should_plot_ci_post_cti=True,
        should_plot_chi_squared_map=True,
        output_path=ci_fit_plotter_path, output_format='png')

    assert ci_fit_plotter_path + 'fit_image.png' in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_noise_map.png' not in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_signal_to_noise_map.png' not in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_ci_pre_cti.png' in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_ci_post_cti.png' in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_residual_map.png' not in plot_patch.paths

    assert ci_fit_plotter_path + 'fit_chi_squared_map.png' in plot_patch.paths



import os

from autofit import conf
from autofit.mapper import prior_model
from autofit.optimize import non_linear as nl

from autocti.charge_injection import ci_data
from autocti.charge_injection import ci_pattern
from autocti.charge_injection import ci_hyper
from autocti.model import arctic_params
from autocti.model import arctic_settings
from autocti.pipeline import phase as ph
from autocti.pipeline import pipeline as pl
from test.integration import tools

shape = (36, 36)
ci_regions = [(1, 7, 1, 30), (17, 23, 1, 30)]
normalizations = [84700.0]
frame_geometry = tools.CIQuadGeometryIntegration()

test_type = 'serial'
test_name = 'x1_species_x1_image_no_pool'

path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))
output_path = path + 'output/' + test_type
config_path = path + 'config'
conf.instance = conf.Config(config_path=config_path, output_path=output_path)


def pipeline():

    serial_species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.5)
    serial_ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                     well_fill_beta=0.8, well_fill_gamma=0.0)
    cti_params = arctic_params.ArcticParams(serial_ccd=serial_ccd, serial_species=[serial_species])

    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=1, n_levels=2000,
                                                 charge_injection_mode=True, readout_offset=0)
    cti_settings = arctic_settings.ArcticSettings(serial=serial_settings)

    tools.reset_paths(test_name=test_name, output_path=output_path)
    tools.simulate_integration_quadrant(test_name=test_name, normalizations=normalizations, cti_params=cti_params,
                                        cti_settings=cti_settings)

    pattern = ci_pattern.CIPatternUniform(normalization=normalizations[0], regions=ci_regions)

    data = ci_data.load_ci_data_from_fits(frame_geometry=frame_geometry, ci_pattern=pattern,
                                          ci_image_path=path + '/data/' + test_name + '/ci_image_0.fits',
                                          ci_noise_map_from_single_value=1.0)

    pipeline = make_pipeline(test_name=test_name)
    pipeline.run(ci_datas=[data], cti_settings=cti_settings)


def make_pipeline(test_name):

    class SerialPhase(ph.SerialPhase):

        def pass_priors(self, previous_results):

            self.serial_ccd.well_fill_alpha = 1.0
            self.serial_ccd.well_fill_gamma = 0.0

    phase1 = SerialPhase(optimizer_class=nl.MultiNest,
                         serial_species=[prior_model.PriorModel(arctic_params.Species)], rows=(0,4),
                         serial_ccd=arctic_params.CCD, phase_name="{}/phase1".format(test_name))

    phase1.optimizer.n_live_points = 60
    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.sampling_efficiency = 0.2

    class SerialHyperModelFixedPhase(ph.SerialHyperPhase):

        def pass_priors(self, previous_results):

            self.serial_species = previous_results[0].constant.serial_species
            self.serial_ccd = previous_results[0].constant.serial_ccd

    phase2 = SerialHyperModelFixedPhase(serial_species=[prior_model.PriorModel(arctic_params.Species)],
                                        serial_ccd=arctic_params.CCD,
                                        hyper_noise_scalar_ci_regions=ci_hyper.CIHyperNoiseScalar,
                                        hyper_noise_scalar_serial_trails=ci_hyper.CIHyperNoiseScalar,
                                        optimizer_class=nl.MultiNest, rows=None,
                                        phase_name="{}/phase2".format(test_name))

    class SerialHyperFixedPhase(ph.SerialHyperPhase):

        def pass_priors(self, previous_results):

            self.hyper_noise_scalar_ci_regions = previous_results[1].constant.hyper_noise_scalar_ci_regions
            self.hyper_noise_scalar_serial_trails = previous_results[1].constant.hyper_noise_scalar_serial_trails
            self.serial_species = previous_results[0].variable.serial_species
            self.serial_ccd = previous_results[0].variable.serial_ccd
            self.serial_ccd.well_fill_alpha = 1.0
            self.serial_ccd.well_fill_gamma = 0.0

    phase3 = SerialHyperFixedPhase(optimizer_class=nl.MultiNest, rows=None, phase_name="{}/phase3".format(test_name))

    # For the final CTI model, constant efficiency mode has a tendancy to sample parameter space too fast and infer an
    # inaccurate model. Thus, we turn it off for phase 2.

    phase3.optimizer.const_efficiency_mode = False
    phase3.optimizer.n_live_points = 50
    phase3.optimizer.sampling_efficiency = 0.3

    return pl.Pipeline(phase1, phase2, phase3)


if __name__ == "__main__":
    pipeline()



from autocti.data import util
from autocti.charge_injection import ci_frame
from autocti.charge_injection import ci_pattern
from autocti.charge_injection import ci_data

from autocti.model import arctic_settings
from autocti.model import arctic_params

import os

# This tool allows one to make simulated charge injection imaging data-sets for calibrating serial charge transfer
# inefficiency, which can be used to test example pipelines and investigate CTI modeling on data-sets where the
# 'true' answer is known.

# The 'image_type' and 'model_type' determine the directory the output data folder, e.g:

# The image will be output as '/workspace/data/image_type/model_type/ci_image.fits'.
# The noise-map will be output as '/workspace/data/image_type/model_type/ci_noise_map.fits'.
# The pre cti ci image will be output as '/workspace/data/image_type/model_type/ci_pre_cti.fits'.

ci_data_name = 'ci_x8_images_uniform'
ci_data_model = 'serial_x1_species'

shape = (2316, 2119) # The shape determines the shape of the images simulated.

# Specify the charge injection regions on the CCD, which in this case is 7 equally spaced rectangular blocks.
ci_regions = [(0, 330, 51, shape[1]-20), (330, 2316, 51, shape[1]-20)]

# The normalization of every ci image - this size of this list thus determines how many images are simulated.
normalizations=[100.0, 500.0, 1000.0, 5000.0, 10000.0, 25000.0, 50000.0, 84700.0]

# The frame geometry of the image being simuated.
frame_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
# frame_geometry = ci_frame.QuadGeometryEuclid.bottom_right()
# frame_geometry = ci_frame.QuadGeometryEuclid.top_left()
# frame_geometry = ci_frame.QuadGeometryEuclid.top_right()

# The CTI settings of arCTIc, which models the CCD read-out including CTI. For serial ci data, we include 'charge
# injection mode' which accounts for the fact that every pixel is transferred over the full CCD.
serial_cti_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                        charge_injection_mode=False, readout_offset=0)
cti_settings = arctic_settings.ArcticSettings(serial=serial_cti_settings)

# The CTI model parameters of arCTIc, which includes each trap species density / lifetime and the CCD properties for
# serial charge transfer.
serial_species = arctic_params.Species(trap_density=0.5, trap_lifetime=4.0)
serial_ccd = arctic_params.CCD(well_notch_depth=1.0e-4, well_fill_beta=0.58, well_fill_gamma=0.0, well_fill_alpha=1.0)
cti_params = arctic_params.ArcticParams(serial_species=[serial_species], serial_ccd=serial_ccd)

# Use the ci normalizations and regions to create the ci pattern of every image that is to be simulated.
ci_patterns = ci_pattern.uniform_from_lists(normalizations=normalizations, regions=ci_regions)

# Use the simulate ci patterns to generate the pre-cti charge injection images.
ci_pre_ctis = list(map(lambda ci_pattern :
                       ci_pattern.simulate_ci_pre_cti(shape=shape),
                       ci_patterns))

# Use every ci pattern to simulate a ci image.
ci_datas = list(map(lambda ci_pre_cti, ci_pattern:
                    ci_data.simulate(ci_pre_cti=ci_pre_cti, frame_geometry=frame_geometry,
                                     ci_pattern=ci_pattern, cti_settings=cti_settings, cti_params=cti_params,
                                     read_noise=4.0),
                    ci_pre_ctis, ci_patterns))

# Now, lets output this simulated ccd-data to the data folder. First, we'll get a relative path.
path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))

# Check folders of the image type and model type exists in the data folder for the images to be output. If they don't,
# make them.

ci_data_path = path +'/data/' + ci_data_name

if not os.path.exists(ci_data_path):
    os.makedirs(ci_data_path)

ci_data_path = path +'/data/' + ci_data_name + '/' + ci_data_model

if not os.path.exists(ci_data_path):
    os.makedirs(ci_data_path)

# Now, output every image to the data folder as the filename 'ci_data_#.fits'
list(map(lambda ci_data, index:
         util.numpy_array_to_fits(array=ci_data.image,
                                  file_path=ci_data_path + '/ci_image_' + str(index) + '.fits', overwrite=True),
         ci_datas, range(len(ci_datas))))

# Output every pre-cti image to the data folder as the filename 'ci_pre_cti_#.fits'. This allows the calibration
# pipeline to load these images as the model pre-cti images, which is necessary for non-uniform ci patterns.
list(map(lambda ci_data, index :
         util.numpy_array_to_fits(array=ci_data.ci_pre_cti,
                                  file_path=ci_data_path + '/ci_pre_cti_' + str(index) + '.fits', overwrite=True),
         ci_datas, range(len(ci_datas))))


#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: tests/python/ArcticSettings_test.py

Created on: 02/13/18
Author: James Nightingale
"""

import os
import shutil

import numpy as np
import pytest
from astropy.io import fits

from autocti.model import arctic_settings


@pytest.fixture(name='info_path')
def test_info():
    info_path = "{}/files/settings/info/".format(os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(info_path):
        shutil.rmtree(info_path)

    os.mkdir(info_path)

    return info_path


@pytest.fixture(name='hdr_path')
def test_header_info():
    hdr_path = "{}/files/settings/header_info/".format(os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(hdr_path):
        shutil.rmtree(hdr_path)

    os.mkdir(hdr_path)

    return hdr_path


class TestArcticSettings:

    class TestConstructor:

        def test__sets_up_parameters_with_correct_values(self):

            parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                       readout_offset=0)

            arctic_both = arctic_settings.ArcticSettings(neomode='NEO', parallel=parallel_settings,
                                                         serial=serial_settings)

            assert arctic_both.neomode == 'NEO'

            assert arctic_both.parallel.well_depth == 84700
            assert arctic_both.parallel.niter == 1
            assert arctic_both.parallel.express == 5
            assert arctic_both.parallel.n_levels == 2000
            assert arctic_both.parallel.charge_injection_mode is True
            assert arctic_both.parallel.readout_offset == 0

            assert arctic_both.serial.well_depth == 84700
            assert arctic_both.serial.niter == 1
            assert arctic_both.serial.express == 5
            assert arctic_both.serial.n_levels == 2000
            assert arctic_both.serial.charge_injection_mode is False
            assert arctic_both.serial.readout_offset == 0

    class TestFitsHeaderInfo:

        def test__parallel_only__sets_up_header_info_consistent_with_previous_vis_pf(self, hdr_path):
            parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            arctic_parallel = arctic_settings.ArcticSettings(neomode='NEO', parallel=parallel_settings)

            hdu = fits.PrimaryHDU(np.ones((1, 1)), fits.Header())
            hdu.header = arctic_parallel.update_fits_header_info(ext_header=hdu.header)
            hdu.writeto(hdr_path + '/test.fits')

            hdu = fits.open(hdr_path + '/test.fits')
            ext_header = hdu[0].header

            assert ext_header['cte_pite'] == 1
            assert ext_header['cte_pwld'] == 84700
            assert ext_header['cte_pnts'] == 2000

        def test__serial_only__sets_up_header_info_consistent_with_previous_vis_pf(self, hdr_path):
            serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                       charge_injection_mode=False, readout_offset=0)

            arctic_serial = arctic_settings.ArcticSettings(neomode='NEO', serial=serial_settings)

            hdu = fits.PrimaryHDU(np.ones((1, 1)), fits.Header())
            hdu.header = arctic_serial.update_fits_header_info(ext_header=hdu.header)
            hdu.writeto(hdr_path + '/test.fits')

            hdu = fits.open(hdr_path + '/test.fits')
            ext_header = hdu[0].header

            assert ext_header['cte_site'] == 1
            assert ext_header['cte_swld'] == 84700
            assert ext_header['cte_snts'] == 2000

        def test__parallel_and_serial__sets_up_header_info_consistent_with_previous_vis_pf(self, hdr_path):
            parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                       charge_injection_mode=False, readout_offset=0)

            arctic_both = arctic_settings.ArcticSettings(neomode='NEO', parallel=parallel_settings,
                                                         serial=serial_settings)

            hdu = fits.PrimaryHDU(np.ones((1, 1)), fits.Header())
            hdu.header = arctic_both.update_fits_header_info(ext_header=hdu.header)
            hdu.writeto(hdr_path + '/test.fits')

            hdu = fits.open(hdr_path + '/test.fits')
            ext_header = hdu[0].header

            assert ext_header['cte_pite'] == 1
            assert ext_header['cte_pwld'] == 84700
            assert ext_header['cte_pnts'] == 2000

            assert ext_header['cte_site'] == 1
            assert ext_header['cte_swld'] == 84700
            assert ext_header['cte_snts'] == 2000



import logging

from autofit.tools.phase import ResultsCollection

logger = logging.getLogger(__name__)


class Pipeline(object):

    def __init__(self, *phases):
        """

        Parameters
        ----------
        phases: [ph.Phase]
            Phases
        """
        self.phases = phases

    def run(self, ci_datas, cti_settings, pool=None):
        results = []
        for i, phase in enumerate(self.phases):
            logger.info("Running Phase {} (Number {})".format(phase.phase_name, i))
            results.append(phase.run(ci_datas=ci_datas, cti_settings=cti_settings,
                                     previous_results=ResultsCollection(results), pool=pool))
        return results

    def __add__(self, other):
        """
        Compose two Pipelines

        Parameters
        ----------
        other: Pipeline
            Another pipeline

        Returns
        -------
        composed_pipeline: Pipeline
            A pipeline that runs all the  phases from this pipeline and then all the phases from the other pipeline
        """
        return Pipeline(*(self.phases + other.phases))



#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: tests/python/ImageFrame_test.py

Created on: 02/13/18
Author: James Nightingale
"""

import os

import numpy as np
import pytest

from autocti import exc
from autocti.charge_injection import ci_frame
from autocti.data import cti_image


@pytest.fixture(scope='function')
def quadrant_data():
    quadrant_data = np.array([[9, 0, 0],
                              [1, 1, 14],
                              [25, -6, 2]])

    return quadrant_data


@pytest.fixture(scope='class')
def image_data():
    image_data = np.array([[1, 0, 0],
                           [1, 1, 1],
                           [0, 1, 0]])

    return image_data


@pytest.fixture(scope='class')
def euclid_data():
    euclid_data = np.zeros((2048, 2066))
    return euclid_data


path = '{}/'.format(os.path.dirname(os.path.realpath(__file__)))


class TestImageFrame:
    class TestConstructor:

        def test__geometry_is_bottom_left__loads_data_and_dimensions(self, euclid_data):
            image = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), array=euclid_data)

            assert type(image.frame_geometry) == ci_frame.QuadGeometryEuclid
            assert image.shape == (2048, 2066)
            assert (image == np.zeros((2048, 2066))).all()

        def test__geometry_is_bottom_right__loads_data_and_dimensions(self, euclid_data):
            image = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(), array=euclid_data)

            assert type(image.frame_geometry) == ci_frame.QuadGeometryEuclid
            assert image.shape == (2048, 2066)
            assert (image == np.zeros((2048, 2066))).all()

        def test__geometry_is_top_left__loads_data_and_dimensions(self, euclid_data):
            image = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=euclid_data)

            assert type(image.frame_geometry) == ci_frame.QuadGeometryEuclid
            assert image.shape == (2048, 2066)
            assert (image == np.zeros((2048, 2066))).all()

        def test__geometry_is_top_right__loads_data_and_dimensions(self, euclid_data):
            image = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=euclid_data)

            assert type(image.frame_geometry) == ci_frame.QuadGeometryEuclid
            assert image.shape == (2048, 2066)
            assert (image == np.zeros((2048, 2066))).all()


class TestQuadrantEuclidGeometry:
    class TestFromCCDID:
        class TestLeftSide:

            def test__ccd_on_left_side_row_1__quadrant_id_E__chooses_bottom_left_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text1', quad_id='E')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_left_side_row_2__quadrant_id_E__chooses_bottom_left_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text2', quad_id='E')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_left_side_row_3__quadrant_id_E__chooses_bottom_left_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text3', quad_id='E')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_left_side_row_1__quadrant_id_F__chooses_bottom_right_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text1', quad_id='F')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_left_side_row_2__quadrant_id_F__chooses_bottom_right_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text2', quad_id='F')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_left_side_row_3__quadrant_id_F__chooses_bottom_right_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text3', quad_id='F')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_left_side_row_1__quadrant_id_G__chooses_top_right_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text1', quad_id='G')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_left_side_row_2__quadrant_id_G__chooses_top_right_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text2', quad_id='G')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_left_side_row_3__quadrant_id_G__chooses_top_right_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text3', quad_id='G')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_left_side_row_1__quadrant_id_H__chooses_top_left_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text1', quad_id='H')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_left_side_row_2__quadrant_id_H__chooses_top_left_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text2', quad_id='H')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_left_side_row_3__quadrant_id_H__chooses_top_left_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text3', quad_id='H')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

        class TestRightSide:

            def test__ccd_on_right_side_row_4__quadrant_id_E__chooses_top_right_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text4', quad_id='E')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_right_side_row_5__quadrant_id_E__chooses_top_right_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text5', quad_id='E')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_right_side_row_6__quadrant_id_E__chooses_top_right_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text6', quad_id='E')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_right_side_row_4__quadrant_id_F__chooses_top_left_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text4', quad_id='F')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_right_side_row_5__quadrant_id_F__chooses_top_left_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text5', quad_id='F')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_right_side_row_6__quadrant_id_F__chooses_top_left_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text6', quad_id='F')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_right_side_row_4__quadrant_id_G__chooses_bottom_left_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text4', quad_id='G')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_right_side_row_5__quadrant_id_G__chooses_bottom_left_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text5', quad_id='G')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_right_side_row_6__quadrant_id_G__chooses_bottom_left_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text6', quad_id='G')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_right_side_row_4__quadrant_id_H__chooses_bottom_right_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text4', quad_id='H')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_right_side_row_5__quadrant_id_H__chooses_bottom_right_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text5', quad_id='H')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid

            def test__ccd_on_right_side_row_6__quadrant_id_H__chooses_bottom_right_quadrant(self):
                quadrant = ci_frame.QuadGeometryEuclid.from_ccd_and_quadrant_id(ccd_id='text6', quad_id='H')

                assert type(quadrant) == ci_frame.QuadGeometryEuclid


class TestQuadrantGeometryEuclidBottomLeft:
    class TestConstrutor:

        def test__sets_up_quadrant__including_correct_scans(self):
            quadrant = ci_frame.QuadGeometryEuclid.bottom_left()

            assert type(quadrant) == ci_frame.QuadGeometryEuclid

            assert quadrant.parallel_overscan == (2066, 2086, 51, 2099)
            assert quadrant.serial_prescan == (0, 2086, 0, 51)
            assert quadrant.serial_overscan == (0, 2086, 2099, 2119)

    class TestRotations:

        def test__rotate_for_parallel_clocking__oriented_as_described_in_documentation(self, quadrant_data):
            # Quadrant 0 - Bottom left panel of Euclid CCD - input ci_pre_ctis should not be rotated for parallel cti.

            quadrant = ci_frame.QuadGeometryEuclid.bottom_left()

            quadrant_rotated = quadrant.rotate_for_parallel_cti(image=quadrant_data)
            assert (quadrant_rotated == quadrant_data).all()

        def test__rotate_for_parallel_clocking_and_back_again__returns_to_original_orientation(self, quadrant_data):
            # Quadrant 0 - Bottom left panel of Euclid CCD - input ci_pre_ctis should not be rotated for parallel cti.

            quadrant = ci_frame.QuadGeometryEuclid.bottom_left()
            quadrant_rotated = quadrant.rotate_for_parallel_cti(image=quadrant_data)
            quadrant_rotated_back = quadrant.rotate_for_parallel_cti(image=quadrant_rotated)

            assert (quadrant_rotated_back == quadrant_data).all()

        def test__rotate_for_serial_clocking__oriented_as_described_in_documentation(self, quadrant_data):
            # Quadrant 0 - Bottom left panel of Euclid CCD - input ci_pre_ctis should be rotated 90 degrees for
            # serial cti.

            quadrant = ci_frame.QuadGeometryEuclid.bottom_left()

            quadrant_rotated = quadrant.rotate_before_serial_cti(image_pre_clocking=quadrant_data)
            assert (quadrant_rotated == quadrant_data.T).all()

        def test__rotate_for_serial_clocking_and_back_again__returns_to_original_orientation(self, quadrant_data):
            # Quadrant 0 - Bottom left panel of Euclid CCD - input ci_pre_ctis should be rotated 90 degrees for
            # serial cti.

            quadrant = ci_frame.QuadGeometryEuclid.bottom_left()
            quadrant_rotated = quadrant.rotate_before_serial_cti(image_pre_clocking=quadrant_data)
            quadrant_rotated_back = quadrant.rotate_after_serial_cti(image_post_clocking=quadrant_rotated)

            assert (quadrant_rotated_back == quadrant_data).all()


class TestQuadrantGeometryEuclidBottomRight:
    class TestConstrutor:

        def test__sets_up_quadrant__including_correct_overscans(self):
            quadrant = ci_frame.QuadGeometryEuclid.bottom_right()

            assert type(quadrant) == ci_frame.QuadGeometryEuclid
            assert quadrant.parallel_overscan == (2066, 2086, 20, 2068)
            assert quadrant.serial_prescan == (0, 2086, 2068, 2119)
            assert quadrant.serial_overscan == (0, 2086, 0, 20)

    class TestRotations:

        def test__rotate_for_parallel_clocking__oriented_as_described_in_documentation(self, quadrant_data):
            # Quadrant 1 - Bottom right panel of Euclid CCD - input ci_pre_ctis should not be rotateped for parallel cti.

            quadrant = ci_frame.QuadGeometryEuclid.bottom_right()

            quadrant_rotated = quadrant.rotate_for_parallel_cti(image=quadrant_data)
            assert (quadrant_rotated == quadrant_data).all()

        def test__rotate_for_parallel_clocking_and_back_again__returns_to_original_orientation(self, quadrant_data):
            # Quadrant 1 - Bottom right panel of Euclid CCD - input ci_pre_ctis should not be rotateped for parallel cti.

            quadrant = ci_frame.QuadGeometryEuclid.bottom_right()
            quadrant_rotated = quadrant.rotate_for_parallel_cti(image=quadrant_data)
            quadrant_rotated_back = quadrant.rotate_for_parallel_cti(image=quadrant_rotated)

            assert (quadrant_rotated_back == quadrant_data).all()

        def test__rotate_for_serial_clocking__oriented_as_described_in_documentation(self, quadrant_data):
            quadrant = ci_frame.QuadGeometryEuclid.bottom_right()

            quadrant_rotated = quadrant.rotate_before_serial_cti(image_pre_clocking=quadrant_data)
            assert (quadrant_rotated == np.fliplr(quadrant_data).T).all()

        def test__rotate_for_serial_clocking_and_back_again__returns_to_original_orientation(self, quadrant_data):
            quadrant = ci_frame.QuadGeometryEuclid.bottom_right()

            quadrant_rotated = quadrant.rotate_before_serial_cti(image_pre_clocking=quadrant_data)
            quadrant_rotated_back = quadrant.rotate_after_serial_cti(image_post_clocking=quadrant_rotated)
            assert (quadrant_rotated_back == quadrant_data).all()


class TestQuadrantGeometryEuclidTopLeft:
    class TestConstrutor:

        def test__sets_up_quadrant__including_correct_overscans(self):
            quadrant = ci_frame.QuadGeometryEuclid.top_left()

            assert type(quadrant) == ci_frame.QuadGeometryEuclid
            assert quadrant.parallel_overscan == (0, 20, 51, 2099)
            assert quadrant.serial_prescan == (0, 2086, 0, 51)
            assert quadrant.serial_overscan == (0, 2086, 2099, 2119)

    class TestRotations:

        def test__rotate_for_parallel_clocking__oriented_as_described_in_documentation(self, quadrant_data):
            # Quadrant 2 - top left panel of Euclid CCD - input ci_pre_ctis should be rotateped upside-down for parallel cti

            quadrant = ci_frame.QuadGeometryEuclid.top_left()

            quadrant_rotated = quadrant.rotate_for_parallel_cti(image=quadrant_data)
            assert (quadrant_rotated == np.flipud(quadrant_data)).all()

        def test__rotate_for_parallel_clocking_and_back_again__returns_to_original_orientation(self, quadrant_data):
            # Quadrant 2 - top left panel of Euclid CCD - input ci_pre_ctis should be rotateped upside-down for parallel cti

            quadrant = ci_frame.QuadGeometryEuclid.top_left()

            quadrant_rotated = quadrant.rotate_for_parallel_cti(image=quadrant_data)
            quadrant_rotated_back = quadrant.rotate_for_parallel_cti(image=quadrant_rotated)

            assert (quadrant_rotated_back == quadrant_data).all()

        def test__rotate_for_serial_clocking__oriented_as_described_in_documentation(self, quadrant_data):
            # Quadrant 2 - top left panel of Euclid CCD - input ci_pre_ctis should be rotated 90 degrees for serial cti.

            quadrant = ci_frame.QuadGeometryEuclid.top_left()

            quadrant_rotated = quadrant.rotate_before_serial_cti(image_pre_clocking=quadrant_data)
            assert (quadrant_rotated == quadrant_data.T).all()

        def test__rotate_for_serial_clocking_and_back_again__returns_to_original_orientation(self, quadrant_data):
            # Quadrant 2 - top left panel of Euclid CCD - input ci_pre_ctis should be rotated 90 degrees for serial cti.

            quadrant = ci_frame.QuadGeometryEuclid.top_left()

            quadrant_rotated = quadrant.rotate_before_serial_cti(image_pre_clocking=quadrant_data)
            quadrant_rotated_back = quadrant.rotate_after_serial_cti(image_post_clocking=quadrant_rotated)
            assert (quadrant_rotated_back == quadrant_data).all()


class TestQuadrantGeometryEuclidTopRight:
    class TestConstrutor:

        def test__sets_up_quadrant__including_correct_overscans(self):
            quadrant = ci_frame.QuadGeometryEuclid.top_right()

            assert type(quadrant) == ci_frame.QuadGeometryEuclid
            assert quadrant.parallel_overscan == (0, 20, 20, 2068)
            assert quadrant.serial_prescan == (0, 2086, 2068, 2119)
            assert quadrant.serial_overscan == (0, 2086, 0, 20)

    class TestRotations:

        def test__rotate_for_parallel_clocking__oriented_as_described_in_documentation(self, quadrant_data):
            # Quadrant 3 - top right panel of Euclid CCD - input ci_pre_ctis should be rotateped upside-down for parallel cti.

            quadrant = ci_frame.QuadGeometryEuclid.top_right()

            quadrant_rotated = quadrant.rotate_for_parallel_cti(image=quadrant_data)
            assert (quadrant_rotated == np.flipud(quadrant_data)).all()

        def test__rotate_for_parallel_clocking_and_back_again__returns_to_original_orientation(self, quadrant_data):
            # Quadrant 3 - top right panel of Euclid CCD - input ci_pre_ctis should be rotateped upside-down for parallel cti.

            quadrant = ci_frame.QuadGeometryEuclid.top_right()

            quadrant_rotated = quadrant.rotate_for_parallel_cti(image=quadrant_data)
            quadrant_rotated_back = quadrant.rotate_for_parallel_cti(image=quadrant_rotated)
            assert (quadrant_rotated_back == quadrant_data).all()

        def test__rotate_for_serial_clocking__oriented_as_described_in_documentation(self, quadrant_data):
            # Quadrant 3 - top right panel of Euclid CCD - input ci_pre_ctis should be rotated 270 degrees for serial cti.

            quadrant = ci_frame.QuadGeometryEuclid.top_right()

            quadrant_rotated = quadrant.rotate_before_serial_cti(image_pre_clocking=quadrant_data)
            assert (quadrant_rotated == np.fliplr(quadrant_data).T).all()

        def test__rotate_for_serial_clocking_and_back_again__returns_to_original_orientation(self, quadrant_data):
            # Quadrant 3 - top right panel of Euclid CCD - input ci_pre_ctis should be rotated 270 degrees for serial cti.

            quadrant = ci_frame.QuadGeometryEuclid.top_right()

            quadrant_rotated = quadrant.rotate_before_serial_cti(image_pre_clocking=quadrant_data)
            quadrant_rotated_back = quadrant.rotate_after_serial_cti(image_post_clocking=quadrant_rotated)
            assert (quadrant_rotated_back == quadrant_data).all()


class TestRegion(object):
    class TestConstructor:

        def test__constructor__converts_region_to_cartesians(self):
            region = ci_frame.Region(region=(0, 1, 2, 3))

            assert region == (0, 1, 2, 3)

            assert region.y0 == 0
            assert region.y1 == 1
            assert region.x0 == 2
            assert region.x1 == 3
            assert region.total_rows == 1
            assert region.total_columns == 1

        def test__first_row_or_column_equal_too_or_bigger_than_second__raise_errors(self):
            with pytest.raises(exc.RegionException):
                ci_frame.Region(region=(2, 2, 1, 2))

            with pytest.raises(exc.RegionException):
                ci_frame.Region(region=(2, 1, 2, 2))

            with pytest.raises(exc.RegionException):
                ci_frame.Region(region=(2, 1, 1, 2))

            with pytest.raises(exc.RegionException):
                ci_frame.Region(region=(0, 1, 3, 2))

        def test__negative_coordinates_raise_errors(self):
            with pytest.raises(exc.RegionException):
                ci_frame.Region(region=(-1, 0, 1, 2))

            with pytest.raises(exc.RegionException):
                ci_frame.Region(region=(0, -1, 1, 2))

            with pytest.raises(exc.RegionException):
                ci_frame.Region(region=(0, 0, -1, 2))

            with pytest.raises(exc.RegionException):
                ci_frame.Region(region=(0, 1, 2, -1))

    class TestExtractRegionFromArray:

        def test__extracts_2x2_region_of_3x3_array(self):
            array = np.array([[1.0, 2.0, 3.0],
                              [4.0, 5.0, 6.0],
                              [7.0, 8.0, 9.0]])

            region = ci_frame.Region(region=(0, 2, 0, 2))

            new_array = array[region.slice]

            assert (new_array == np.array([[1.0, 2.0],
                                           [4.0, 5.0]])).all()

        def test__extracts_2x3_region_of_3x3_array(self):
            array = np.array([[1.0, 2.0, 3.0],
                              [4.0, 5.0, 6.0],
                              [7.0, 8.0, 9.0]])

            region = ci_frame.Region(region=(1, 3, 0, 3))

            new_array = array[region.slice]

            assert (new_array == np.array([[4.0, 5.0, 6.0],
                                           [7.0, 8.0, 9.0]])).all()

    class TestAddRegionToArrayFromImage:

        def test__array_is_all_zeros__image_goes_into_correct_region_of_array(self):
            array = np.zeros((2, 2))
            image = np.ones((2, 2))
            region = ci_frame.Region(region=(0, 1, 0, 1))

            array[region.slice] += image[region.slice]

            assert (array == np.array([[1.0, 0.0],
                                       [0.0, 0.0]])).all()

        def test__array_is_all_1s__image_goes_into_correct_region_of_array_and_adds_to_it(self):
            array = np.ones((2, 2))
            image = np.ones((2, 2))
            region = ci_frame.Region(region=(0, 1, 0, 1))

            array[region.slice] += image[region.slice]

            assert (array == np.array([[2.0, 1.0],
                                       [1.0, 1.0]])).all()

        def test__different_region(self):
            array = np.ones((3, 3))
            image = np.ones((3, 3))
            region = ci_frame.Region(region=(1, 3, 2, 3))

            array[region.slice] += image[region.slice]

            assert (array == np.array([[1.0, 1.0, 1.0],
                                       [1.0, 1.0, 2.0],
                                       [1.0, 1.0, 2.0]])).all()

    class TestSetRegionToZeros:

        def test__region_is_corner__sets_to_0(self):
            array = np.ones((2, 2))

            region = ci_frame.Region(region=(0, 1, 0, 1))

            array[region.slice] = 0

            assert (array == np.array([[0.0, 1.0],
                                       [1.0, 1.0]])).all()

        def test__different_region___sets_to_0(self):
            array = np.ones((3, 3))

            region = ci_frame.Region(region=(1, 3, 2, 3))

            array[region.slice] = 0

            assert (array == np.array([[1.0, 1.0, 1.0],
                                       [1.0, 1.0, 0.0],
                                       [1.0, 1.0, 0.0]])).all()



#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: python/VIS_CTI_Image/ImageFrame.py

Created on: 02/13/18
Author: James Nightingale
"""

import numpy as np

from autocti.data import util


class ImageFrame(np.ndarray):

    def __new__(cls, frame_geometry, array, **kwargs):
        """The CCD ci_frame of an image, including its geometry and therefore the directions parallel and serial CTI are \
        defined.

        Parameters
        ----------
        frame_geometry : ImageFrame.FrameGeometry
            The geometry of the ci_frame, defining the direction of parallel and serial clocking and the \
            locations of different regions of the CCD (overscans, prescan, etc.)
        array : ndarray
            The 2D array of the ci_data of this ci_frame.
        """
        quad = np.array(array, dtype='float64').view(cls)
        quad.frame_geometry = frame_geometry
        return quad

    # noinspection PyUnusedLocal
    def __init__(self, frame_geometry, array):
        """
        Params
        ----------
        frame_geometry : ImageFrame.FrameGeometry
            The geometry of the ci_frame, defining the direction of parallel and serial clocking and the \
            locations of different regions of the CCD (overscans, prescan, etc.)
        array : ndarray
            The 2D array of the ci_data of this ci_frame.
        """
        # noinspection PyArgumentList
        super(ImageFrame, self).__init__()
        self.frame_geometry = frame_geometry

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(ImageFrame, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        class_dict = {}
        for key, value in self.__dict__.items():
            class_dict[key] = value
        new_state = pickled_state[2] + (class_dict,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return pickled_state[0], pickled_state[1], new_state

    # noinspection PyMethodOverriding
    def __setstate__(self, state):

        for key, value in state[-1].items():
            setattr(self, key, value)
        super(ImageFrame, self).__setstate__(state[0:-1])

    @classmethod
    def from_fits(cls, file_path, hdu, frame_geometry):
        """Load the image ci_data from a fits file.

        Params
        ----------
        path : str
            The path to the ci_data
        filename : str
            The file phase_name of the fits image ci_data.
        hdu : int
            The HDU number in the fits file containing the image ci_data.
        frame_geometry : ImageFrame.FrameGeometry
            The geometry of the ci_frame, defining the direction of parallel and serial clocking and the \
            locations of different regions of the CCD (overscans, prescan, etc.)
        """
        return cls(frame_geometry=frame_geometry, array=util.numpy_array_from_fits(file_path=file_path, hdu=hdu))

    def output_as_fits(self, file_path, overwrite=False):
        """Output the image ci_data as a fits file.

        Params
        ----------
        path : str
            The output nlo path of the ci_data
        filename : str
            The file phase_name of the output image.
        """
        util.numpy_array_to_fits(array=self, file_path=file_path, overwrite=overwrite)

    def add_cti_to_image(self, cti_params, cti_settings):
        """Add cti to the image.

        Parameters
        ----------
        cti_params : ArcticParams.ArcticParams
            The CTI model parameters (trap density, trap lifetimes etc.).
        cti_settings : ArcticSettings.ArcticSettings
            The settings that control the cti clocking algorithm (e.g. ccd well_depth express option).
        """
        return self.frame_geometry.add_cti(image=self, cti_params=cti_params, cti_settings=cti_settings)

    def correct_cti_from_image(self, cti_params, cti_settings):
        """Correct cti from the image.

        Parameters
        ----------
        cti_params : ArcticParams.ArcticParams
            The CTI model parameters (trap density, trap lifetimes etc.).
        cti_settings : ArcticSettings.ArcticSettings
            The settings that control the cti clocking algorithm (e.g. ccd well_depth express option).
        """
        return self.frame_geometry.correct_cti(image=self, cti_params=cti_params, cti_settings=cti_settings)



import numpy as np
import pytest

from autocti.data import cti_image
from autocti.charge_injection import ci_frame
from autocti.model import arctic_params
from autocti.model import arctic_settings


def setup(include_parallel=False, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
          p_well_notch_depth=0.01, p_well_fill_alpha=1.0, p_well_fill_beta=0.8,
          p_well_fill_gamma=0.0, include_serial=False, s_trap_densities=(0.05, 0.05,),
          s_trap_lifetimes=(1.0, 1.0),
          s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
          s_well_fill_gamma=0.0):
    serial_species = []
    parallel_species = []
    serial_ccd = None
    parallel_ccd = None

    if include_serial:
        serial_ccd = arctic_params.CCD(well_notch_depth=s_well_notch_depth, well_fill_alpha=s_well_fill_alpha,
                                       well_fill_beta=s_well_fill_beta, well_fill_gamma=s_well_fill_gamma)
        serial_species = [arctic_params.Species(trap_density=s_trap_densities[i], trap_lifetime=s_trap_lifetimes[i]) for
                          i in range(len(s_trap_densities))]
    if include_parallel:
        parallel_ccd = arctic_params.CCD(well_notch_depth=p_well_notch_depth, well_fill_alpha=p_well_fill_alpha,
                                         well_fill_beta=p_well_fill_beta, well_fill_gamma=p_well_fill_gamma)
        parallel_species = [arctic_params.Species(trap_density=p_trap_densities[i], trap_lifetime=p_trap_lifetimes[i])
                            for
                            i in range(len(p_trap_densities))]

    return arctic_params.ArcticParams(parallel_ccd=parallel_ccd, serial_ccd=serial_ccd, serial_species=serial_species,
                                      parallel_species=parallel_species)


@pytest.fixture(scope='class', name='arctic_parallel')
def make_arctic_parallel():
    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                 readout_offset=0)
    arctic_parallel = arctic_settings.ArcticSettings(neomode='NEO', parallel=parallel_settings)

    return arctic_parallel


@pytest.fixture(scope='class', name='arctic_serial')
def make_arctic_serial():
    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                               readout_offset=0)

    arctic_serial = arctic_settings.ArcticSettings(neomode='NEO', serial=serial_settings)

    return arctic_serial


@pytest.fixture(scope='class', name='arctic_both')
def make_arctic_both():
    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                 readout_offset=0)

    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                               readout_offset=0)

    arctic_both = arctic_settings.ArcticSettings(neomode='NEO', parallel=parallel_settings,
                                                 serial=serial_settings)

    return arctic_both


@pytest.fixture(scope='class', name='params_parallel')
def make_params_parallel():
    params_parallel = [arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)]
    ccd = arctic_params.CCD(well_notch_depth=0.000001, well_fill_beta=0.8)

    params_parallel = arctic_params.ArcticParams(parallel_species=params_parallel, parallel_ccd=ccd)

    return params_parallel


@pytest.fixture(scope='class', name='params_serial')
def make_params_serial():
    params_serial = [arctic_params.Species(trap_density=0.2, trap_lifetime=2.0)]
    ccd = arctic_params.CCD(well_notch_depth=0.000001, well_fill_beta=0.4)

    params_serial = arctic_params.ArcticParams(serial_species=params_serial, serial_ccd=ccd)

    return params_serial


@pytest.fixture(scope='class', name='params_both')
def make_params_both():
    params_parallel = [arctic_params.Species(trap_density=0.4, trap_lifetime=1.0)]
    parallel_ccd = arctic_params.CCD(well_notch_depth=0.000001, well_fill_beta=0.8)

    params_serial = [arctic_params.Species(trap_density=0.2, trap_lifetime=2.0)]
    serial_ccd = arctic_params.CCD(well_notch_depth=0.000001, well_fill_beta=0.4)

    params_both = arctic_params.ArcticParams(parallel_species=params_parallel, serial_species=params_serial,
                                             parallel_ccd=parallel_ccd, serial_ccd=serial_ccd)

    return params_both


class TestQuadrantGeometryEuclidBottomLeft:
    class TestAddCTI:
        class TestParallelCTI:

            def test__horizontal_charge_line__loses_charge_and_trails_form_below_line(self, arctic_parallel,
                                                                                      params_parallel):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, :] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_parallel, cti_settings=arctic_parallel)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[0:2, :] == 0.0).all()  # No trails closest to read out
                assert (image_difference[2, :] < 0.0).all()  # line loses charge
                assert (image_difference[3:5, :] > 0.0).all()  # trails appear below line

            def test__vertical_charge_line__loses_charge_no_trails(self, arctic_parallel, params_parallel):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[:, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_parallel, cti_settings=arctic_parallel)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[:, 0:2] == 0.0).all()  # No trails closest to read out
                assert (image_difference[:, 3:-1] == 0.0).all()  # line loses charge
                assert (image_difference[:, 2] < 0.0).all()  # No trails away from read-out either, as wrong direction

        class TestSerialCTI:

            def test__horizontal_charge_line__loses_charge_no_trails(self, arctic_serial, params_serial):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, :] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_serial, cti_settings=arctic_serial)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[0:2, :] == 0.0).all()  # No trails closest to read out
                assert (image_difference[2, :] < 0.0).all()  # line loses charge
                assert (image_difference[3:5,
                        :] == 0.0).all()  # No trails away from read-out either, as wrong direction

            def test__vertical_charge_line__loses_charge_and_trails_form_to_right_of_line(self, arctic_serial,
                                                                                          params_serial):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[:, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_serial, cti_settings=arctic_serial)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[:, 0:2] == 0.0).all()  # No trails closest to read out
                assert (image_difference[:, 2] < 0.0).all()  # line loses charge
                assert (image_difference[:, 3:5] > 0.0).all()  # Trails appear away from read-out

        class TestParallelAndSerialCTI:

            def test__horizontal_charge_line__loses_charge_trails_form_both_directions(self, arctic_both,
                                                                                       params_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 1:4] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[0:2, :] == 0.0).all()  # No change in front of charge
                assert (image_difference[2, 1:4] < 0.0).all()  # charge lost in charge

                assert (image_difference[3:5, 1:4] > 0.0).all()  # Parallel trails behind charge

                assert (image_difference[2:5, 0] == 0.0).all()  # no serial cti trail to left
                assert (image_difference[2:5, 4] > 0.0).all()  # serial cti trail to right including parallel cti trails

                assert (image_difference[3, 1:4] > image_difference[4, 1:4]).all()
                # check parallel cti trails decreasing.

                assert (image_difference[3, 4] > image_difference[
                    4, 4])  # Check serial trails of paralle trails decreasing.

            def test__vertical_charge_line__loses_charge_trails_form_in_serial_directions(self, arctic_both,
                                                                                          params_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[1:4, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[0, 0:5] == 0.0).all()  # No change in front of charge
                assert (image_difference[1:4, 2] < 0.0).all()  # charge lost in charge

                assert (image_difference[4, 2] > 0.0).all()  # Parallel trail behind charge

                assert (image_difference[0:5, 0:2] == 0.0).all()  # no serial cti trail to left
                assert (image_difference[1:5,
                        3:5] > 0.0).all()  # serial cti trail to right including parallel cti trails

                assert (image_difference[3, 3] > image_difference[3, 4])  # Check serial trails decreasing.
                assert (image_difference[4, 3] > image_difference[
                    4, 4])  # Check serial trails of parallel trails decreasing.

            def test__individual_pixel_trails_form_cross_around_it(self, params_both, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[0:2, :] == 0.0).all()  # First two rows should all remain zero
                assert (image_difference[:, 0:2] == 0.0).all()  # First tow columns should all remain zero
                assert (image_difference[2, 2] < 0.0)  # pixel which had charge should lose it due to cti.
                assert (image_difference[3:5, 2] > 0.0).all()  # Parallel trail increases charge above pixel
                assert (image_difference[2, 3:5] > 0.0).all()  # Serial trail increases charge to right of pixel
                assert (image_difference[3:5,
                        3:5] > 0.0).all()  # Serial trailing of parallel trail increases charge up-right of pixel

            def test__individual_pixel_double_density__more_captures_so_brighter_trails(self, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), array=image_pre_cti)

                params_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_0 = im.add_cti_to_image(cti_params=params_0, cti_settings=arctic_both)

                params_1 = setup(include_parallel=True, p_trap_densities=(0.2,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.2,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_1 = im.add_cti_to_image(cti_params=params_1, cti_settings=arctic_both)

                image_difference = image_post_cti_1 - image_post_cti_0

                assert (image_difference[0:2, :] == 0.0).all()  # First two rows remain zero
                assert (image_difference[:, 0:2] == 0.0).all()  # First tow columns remain zero
                assert (image_difference[
                            2, 2] < 0.0)  # More captures in second ci_pre_ctis, so more charge in central pixel lost
                assert (image_difference[3:5,
                        2] > 0.0).all()  # More captpures in ci_pre_ctis 2, so brighter parallel trail
                assert (image_difference[2,
                        3:5] > 0.0).all()  # More captpures in ci_pre_ctis 2, so brighter serial trail
                assert (image_difference[3:5, 3:5] > 0.0).all()  # Brighter serial trails from parallel trail trails

            def test__individual_pixel_increase_lifetime_longer_release_so_fainter_trails(self, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), array=image_pre_cti)

                params_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_0 = im.add_cti_to_image(cti_params=params_0, cti_settings=arctic_both)

                params_1 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(20.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(20.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_1 = im.add_cti_to_image(cti_params=params_1, cti_settings=arctic_both)

                image_difference = image_post_cti_1 - image_post_cti_0

                assert (image_difference[0:2, :] == 0.0).all()  # First two rows remain zero
                assert (image_difference[:, 0:2] == 0.0).all()  # First tow columns remain zero
                assert (image_difference[2, 2] == 0.0)  # Same density so same captures
                assert (image_difference[3:5,
                        2] < 0.0).all()  # Longer release in ci_pre_ctis 2, so fainter parallel trail
                assert (image_difference[2,
                        3:5] < 0.0).all()  # Longer release in ci_pre_ctis 2, so fainter serial trail
                assert (image_difference[3:5,
                        3:5] < 0.0).all()  # Longer release in ci_pre_ctis 2, so fainter parallel trail trails

            def test__individual_pixel_increase_beta__fewer_captures_so_fainter_trails(self, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), array=image_pre_cti)

                params_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_0 = im.add_cti_to_image(cti_params=params_0, cti_settings=arctic_both)

                params_1 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.9,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.9)

                image_post_cti_1 = im.add_cti_to_image(cti_params=params_1, cti_settings=arctic_both)

                image_difference = image_post_cti_1 - image_post_cti_0

                assert (image_difference[0:2, :] == 0.0).all()  # First two rows remain zero
                assert (image_difference[:, 0:2] == 0.0).all()  # First tow columns remain zero
                assert (image_difference[2, 2] > 0.0)  # Higher beta in 2, so fewer captures
                assert (image_difference[3:5, 2] < 0.0).all()  # Fewer catprues in 2, so fainter parallel trail
                assert (image_difference[2, 3:5] < 0.0).all()  # Fewer captures in 2, so fainter serial trail
                assert (image_difference[3:5, 3:5] < 0.0).all()  # fewer captures in 2, so fainter trails trail region

    class TestCorrectCTI:
        class TestParallelCTI:

            def test__array_of_values__corrected_image_more_like_original(self, arctic_parallel, params_parallel):
                image_pre_cti = np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 3.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 352, 9.4, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 0.0, 9.0, 0.0],
                                          [0.0, 9.0, 9.1, 9.3, 9.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                        array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_parallel, cti_settings=arctic_parallel)
                image_difference_1 = image_post_cti - image_pre_cti

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                        array=image_post_cti)

                image_correct_cti = im.correct_cti_from_image(cti_params=params_parallel, cti_settings=arctic_parallel)
                image_difference_2 = image_correct_cti - image_pre_cti

                assert (abs(image_difference_2) <= abs(image_difference_1)).all()

        class TestSerialCTI:

            def test__array_of_values__corrected_image_more_like_original(self, arctic_serial, params_serial):
                image_pre_cti = np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 3.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 352, 9.4, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 0.0, 9.0, 0.0],
                                          [0.0, 9.0, 9.1, 9.3, 9.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_serial, cti_settings=arctic_serial)
                image_difference_1 = image_post_cti - image_pre_cti

                # Update ci_pre_ctis ci_data so that it corrects the cti added ci_pre_ctis
                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                        array=image_post_cti)

                image_correct_cti = im.correct_cti_from_image(cti_params=params_serial, cti_settings=arctic_serial)
                image_difference_2 = image_correct_cti - image_pre_cti

                assert (abs(image_difference_2) <= abs(image_difference_1)).all()


class TestQuadrantGeometryEuclidBottomRight:
    class TestAddCTI:
        class TestParallelCTI:

            def test__horizontal_charge_line__loses_charge_and_trails_form_below(self, arctic_parallel,
                                                                                 params_parallel):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, :] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_parallel, cti_settings=arctic_parallel)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[0:2, :] == 0.0).all()  # No trails closest to read out
                assert (image_difference[2, :] < 0.0).all()  # line loses charge
                assert (image_difference[3:5, :] > 0.0).all()  # Trails appear away from read-out

            def test__vertical_charge_line__loses_charge_no_trails(self, arctic_parallel, params_parallel):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[:, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_parallel, cti_settings=arctic_parallel)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[:, 0:2] == 0.0).all()  # No trails closest to read out
                assert (image_difference[:, 2] < 0.0).all()  # line loses charge
                assert (image_difference[:,
                        3:5] == 0.0).all()  # No trails away from read-out either, as wrong direction

        class TestSerialCTI:

            def test__horizontal_charge_line__loses_charge_no_trails(self, arctic_serial, params_serial):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, :] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_serial, cti_settings=arctic_serial)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[0:2, :] == 0.0).all()  # No trails closest to read out
                assert (image_difference[2, :] < 0.0).all()  # line loses charge
                assert (image_difference[3:5,
                        :] == 0.0).all()  # No trails away from read-out either, as wrong direction

            def test__vertical_charge_line__loses_charge_and_trails_form_to_left_of_line(self, arctic_serial,
                                                                                         params_serial):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[:, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_serial, cti_settings=arctic_serial)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[:, 0:2] > 0.0).all()  # No trails closest to read out
                assert (image_difference[:, 2] < 0.0).all()  # line loses charge
                assert (image_difference[:, 3:5] == 0.0).all()  # Trails appear away from read-out

        class TestParallelAndSerialCTI:

            def test__horizontal_charge_line__loses_charge_trails_form_both_directions(self, arctic_both,
                                                                                       params_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 1:4] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[0:2, :] == 0.0).all()  # No change in front of charge
                assert (image_difference[2, 1:4] < 0.0).all()  # charge lost in charge

                assert (image_difference[3:5, 1:4] > 0.0).all()  # Parallel trails behind charge

                assert (image_difference[2:5, 0] > 0.0).all()  # serial cti trail to left including parallel trails
                assert (image_difference[2:5, 4] == 0.0).all()  # no trails to right

                assert (image_difference[3, 1:4] > image_difference[4, 1:4]).all()  # check parallel cti trails
                # decreasing.

                assert (image_difference[3, 0] > image_difference[
                    4, 0])  # Check serial trails of parallel trails decreasing.

            def test__vertical_charge_line__loses_charge_trails_form_in_serial_direction(self, arctic_both,
                                                                                         params_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[1:4, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[0, 0:5] == 0.0).all()  # No change in front of charge
                assert (image_difference[1:4, 2] < 0.0).all()  # charge lost in charge

                assert (image_difference[4, 2] > 0.0).all()  # Parallel trail behind charge

                assert (image_difference[1:5,
                        0:2] > 0.0).all()  # serial cti trail is to the left including parallel trails
                assert (image_difference[0:5, 3:5] == 0.0).all()  # no serial cti trail to the right

                assert (image_difference[3, 1] > image_difference[3, 0])  # Check serial trails decreasing.
                assert (image_difference[4, 1] > image_difference[
                    4, 0])  # Check serial trails of parallel trails decreasing.

            def test__individual_pixel_trails_form_cross_around_it(self, params_both, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[0:2, :] == 0.0).all()  # First two rows should all remain zero
                assert (image_difference[:, 3:5] == 0.0).all()  # First tow columns should all remain zero
                assert (image_difference[2, 2] < 0.0)  # pixel which had charge should lose it due to cti.
                assert (image_difference[3:5, 2] > 0.0).all()  # Parallel trail increases charge above pixel
                assert (image_difference[2, 0:2] > 0.0).all()  # Serial trail increases charge to right of pixel
                assert (image_difference[3:5,
                        0:2] > 0.0).all()  # Serial trailing of parallel trail increases charge up-right of pixel

            def test__individual_pixel_double_density__more_captures_so_brighter_trails(self, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_pre_cti)

                params_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_0 = im.add_cti_to_image(cti_params=params_0, cti_settings=arctic_both)

                params_1 = setup(include_parallel=True, p_trap_densities=(0.2,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.2,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_1 = im.add_cti_to_image(cti_params=params_1, cti_settings=arctic_both)

                image_difference = image_post_cti_1 - image_post_cti_0

                assert (image_difference[0:2, :] == 0.0).all()  # First two rows remain zero
                assert (image_difference[:, 3:5] == 0.0).all()  # First tow columns remain zero
                assert (image_difference[
                            2, 2] < 0.0)  # More captures in second ci_pre_ctis, so more charge in central pixel lost
                assert (image_difference[3:5,
                        2] > 0.0).all()  # More captpures in ci_pre_ctis 2, so brighter parallel trail
                assert (image_difference[2,
                        0:2] > 0.0).all()  # More captpures in ci_pre_ctis 2, so brighter serial trail
                assert (image_difference[3:5, 0:2] > 0.0).all()  # Brighter serial trails from parallel trail trails

            def test__individual_pixel_increase_lifetime_longer_release_so_fainter_trails(self, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_pre_cti)

                params_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_0 = im.add_cti_to_image(cti_params=params_0, cti_settings=arctic_both)

                params_1 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(20.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(20.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_1 = im.add_cti_to_image(cti_params=params_1, cti_settings=arctic_both)

                image_difference = image_post_cti_1 - image_post_cti_0

                assert (image_difference[0:2, :] == 0.0).all()  # First two rows remain zero
                assert (image_difference[:, 3:5] == 0.0).all()  # First tow columns remain zero
                assert (image_difference[2, 2] == 0.0)  # Same density so same captures
                assert (image_difference[3:5,
                        2] < 0.0).all()  # Longer release in ci_pre_ctis 2, so fainter parallel trail
                assert (image_difference[2,
                        0:2] < 0.0).all()  # Longer release in ci_pre_ctis 2, so fainter serial trail
                assert (image_difference[3:5,
                        0:2] < 0.0).all()  # Longer release in ci_pre_ctis 2, so fainter parallel trail trails

            def test__individual_pixel_increase_beta__fewer_captures_so_fainter_trails(self, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_pre_cti)

                params_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_0 = im.add_cti_to_image(cti_params=params_0, cti_settings=arctic_both)

                params_1 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.9,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.9)

                image_post_cti_1 = im.add_cti_to_image(cti_params=params_1, cti_settings=arctic_both)

                image_difference = image_post_cti_1 - image_post_cti_0

                assert (image_difference[0:2, :] == 0.0).all()  # First two rows remain zero
                assert (image_difference[:, 3:5] == 0.0).all()  # First tow columns remain zero
                assert (image_difference[2, 2] > 0.0)  # Higher beta in 2, so fewer captures
                assert (image_difference[3:5, 2] < 0.0).all()  # Fewer catprues in 2, so fainter parallel trail
                assert (image_difference[2, 0:2] < 0.0).all()  # Fewer captures in 2, so fainter serial trail
                assert (image_difference[3:5, 0:2] < 0.0).all()  # fewer captures in 2, so fainter trails trail region

    class TestCorrectCTI:
        class TestParallelCTI:

            def test__array_of_values__corrected_image_more_like_original(self, arctic_parallel, params_parallel):
                image_pre_cti = np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 3.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 352, 9.4, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 0.0, 9.0, 0.0],
                                          [0.0, 9.0, 9.1, 9.3, 9.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_parallel, cti_settings=arctic_parallel)
                image_difference_1 = image_post_cti - image_pre_cti

                # Update ci_pre_ctis ci_data so that it corrects the cti added ci_pre_ctis
                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_post_cti)

                image_correct_cti = im.correct_cti_from_image(cti_params=params_parallel, cti_settings=arctic_parallel)
                image_difference_2 = image_correct_cti - image_pre_cti

                assert (abs(image_difference_2) <= abs(image_difference_1)).all()

        class TestSerialCTI:

            def test__array_of_values__corrected_image_more_like_original(self, arctic_serial, params_serial):
                image_pre_cti = np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 3.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 352, 9.4, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 0.0, 9.0, 0.0],
                                          [0.0, 9.0, 9.1, 9.3, 9.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_serial, cti_settings=arctic_serial)
                image_difference_1 = image_post_cti - image_pre_cti

                # Update ci_pre_ctis ci_data so that it corrects the cti added ci_pre_ctis
                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(),
                                        array=image_post_cti)

                image_correct_cti = im.correct_cti_from_image(cti_params=params_serial, cti_settings=arctic_serial)
                image_difference_2 = image_correct_cti - image_pre_cti

                assert (abs(image_difference_2) <= abs(image_difference_1)).all()


class TestQuadrantGeometryEuclidTopLeft:
    class TestAddCTI:
        class TestParallelCTI:

            def test__horizontal_charge_line__loses_charge_and_trails_form_above_line(self, arctic_parallel,
                                                                                      params_parallel):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, :] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_parallel, cti_settings=arctic_parallel)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[3:5, :] == 0.0).all()  # No trails closest to read out
                assert (image_difference[2, :] < 0.0).all()  # line loses charge
                assert (image_difference[0:2, :] > 0.0).all()  # trails appear behind line

            def test__vertical_charge_line__loses_charge_no_trails(self, arctic_parallel, params_parallel):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[:, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_parallel, cti_settings=arctic_parallel)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[:, 0:2] == 0.0).all()  # No trails closest to read out
                assert (image_difference[:, 3:-1] == 0.0).all()  # line loses charge
                assert (image_difference[:, 2] < 0.0).all()  # No trails away from read-out either, as wrong direction

        class TestSerialCTI:

            def test__horizontal_charge_line__loses_charge_no_trails(self, arctic_serial, params_serial):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, :] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_serial, cti_settings=arctic_serial)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[0:2, :] == 0.0).all()  # No trails closest to read out
                assert (image_difference[2, :] < 0.0).all()  # line loses charge
                assert (image_difference[3:5,
                        :] == 0.0).all()  # No trails away from read-out either, as wrong direction

            def test__vertical_charge_line__loses_charge_and_trails_form_to_right_of_line(self, arctic_serial,
                                                                                          params_serial):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[:, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_serial, cti_settings=arctic_serial)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[:, 0:2] == 0.0).all()  # No trails closest to read out
                assert (image_difference[:, 2] < 0.0).all()  # line loses charge
                assert (image_difference[:, 3:5] > 0.0).all()  # Trails appear away from read-out

        class TestParallelAndSerialCTI:

            def test__horizontal_charge_line__loses_charge_trails_form_both_directions(self, arctic_both,
                                                                                       params_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 1:4] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[3:5, :] == 0.0).all()  # No change in front of charge
                assert (image_difference[2, 1:4] < 0.0).all()  # charge lost in charge

                assert (image_difference[0:2, 1:4] > 0.0).all()  # Parallel trails behind charge

                assert (image_difference[0:5, 0] == 0.0).all()  # no serial cti trail to left
                assert (image_difference[0:3, 4] > 0.0).all()  # serial cti trail to right including parallel cti trails

                assert (image_difference[1, 1:4] > image_difference[0, 1:4]).all()  # check parallel cti trails
                # decreasing.

                assert (image_difference[1, 4] > image_difference[
                    0, 4])  # Check serial trails of parallel trails decreasing.

            def test__vertical_charge_line__loses_charge_trails_form_in_serial_direction(self, arctic_both,
                                                                                         params_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[1:4, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[4, 0:5] == 0.0).all()  # No change in front of charge
                assert (image_difference[1:4, 2] < 0.0).all()  # charge lost in charge

                assert (image_difference[0, 2] > 0.0).all()  # Parallel trail behind charge

                assert (image_difference[1:5, 0:2] == 0.0).all()  # no serial cti trail to left
                assert (image_difference[0:4,
                        3:5] > 0.0).all()  # serial cti trail to right including parallel cti trails

                assert (image_difference[3, 3] > image_difference[3, 4])  # Check serial trails decreasing.
                assert (image_difference[0, 3] > image_difference[
                    0, 4])  # Check serial trails of parallel trails decreasing.

            def test__individual_pixel__trails_form_cross_around_it(self, params_both, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[3:5, :] == 0.0).all()  # First two rows should all remain zero
                assert (image_difference[:, 0:2] == 0.0).all()  # First tow columns should all remain zero
                assert (image_difference[2, 2] < 0.0)  # pixel which had charge should lose it due to cti.
                assert (image_difference[0:2, 2] > 0.0).all()  # Parallel trail increases charge above pixel
                assert (image_difference[2, 3:5] > 0.0).all()  # Serial trail increases charge to right of pixel
                assert (image_difference[0:2,
                        3:5] > 0.0).all()  # Serial trailing of parallel trail increases charge up-right of pixel

            def test__individual_pixel_double_density__more_captures_so_brighter_trails(self, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=image_pre_cti)

                params_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_0 = im.add_cti_to_image(cti_params=params_0, cti_settings=arctic_both)

                params_1 = setup(include_parallel=True, p_trap_densities=(0.2,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.2,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_1 = im.add_cti_to_image(cti_params=params_1, cti_settings=arctic_both)

                image_difference = image_post_cti_1 - image_post_cti_0

                assert (image_difference[3:5, :] == 0.0).all()  # First two rows remain zero
                assert (image_difference[:, 0:2] == 0.0).all()  # First tow columns remain zero
                assert (image_difference[
                            2, 2] < 0.0)  # More captures in second ci_pre_ctis, so more charge in central pixel lost
                assert (image_difference[0:2,
                        2] > 0.0).all()  # More captpures in ci_pre_ctis 2, so brighter parallel trail
                assert (image_difference[2,
                        3:5] > 0.0).all()  # More captpures in ci_pre_ctis 2, so brighter serial trail
                assert (image_difference[0:2, 3:5] > 0.0).all()  # Brighter serial trails from parallel trail trails

            def test__individual_pixel_increase_lifetime_longer_release_so_fainter_trails(self, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=image_pre_cti)

                params_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_0 = im.add_cti_to_image(cti_params=params_0, cti_settings=arctic_both)

                params_1 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(20.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(20.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_1 = im.add_cti_to_image(cti_params=params_1, cti_settings=arctic_both)

                image_difference = image_post_cti_1 - image_post_cti_0

                assert (image_difference[3:5, :] == 0.0).all()  # First two rows remain zero
                assert (image_difference[:, 0:2] == 0.0).all()  # First tow columns remain zero
                assert (image_difference[2, 2] == 0.0)  # Same density so same captures
                assert (image_difference[0:2,
                        2] < 0.0).all()  # Longer release in ci_pre_ctis 2, so fainter parallel trail
                assert (image_difference[2,
                        3:5] < 0.0).all()  # Longer release in ci_pre_ctis 2, so fainter serial trail
                assert (image_difference[0:2,
                        3:5] < 0.0).all()  # Longer release in ci_pre_ctis 2, so fainter parallel trail trails

            def test__individual_pixel_increase_beta__fewer_captures_so_fainter_trails(self, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=image_pre_cti)

                params_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_0 = im.add_cti_to_image(cti_params=params_0, cti_settings=arctic_both)

                params_1 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.9,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.9)

                image_post_cti_1 = im.add_cti_to_image(cti_params=params_1, cti_settings=arctic_both)

                image_difference = image_post_cti_1 - image_post_cti_0

                assert (image_difference[3:5, :] == 0.0).all()  # First two rows remain zero
                assert (image_difference[:, 0:2] == 0.0).all()  # First tow columns remain zero
                assert (image_difference[2, 2] > 0.0)  # Higher beta in 2, so fewer captures
                assert (image_difference[0:2, 2] < 0.0).all()  # Fewer catprues in 2, so fainter parallel trail
                assert (image_difference[2, 3:5] < 0.0).all()  # Fewer captures in 2, so fainter serial trail
                assert (image_difference[0:2, 3:5] < 0.0).all()  # fewer captures in 2, so fainter trails trail region

    class TestCorrectCTI:
        class TestParallelCTI:

            def test__array_of_values__corrected_image_more_like_original(self, arctic_parallel, params_parallel):
                image_pre_cti = np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 3.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 352, 9.4, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 0.0, 9.0, 0.0],
                                          [0.0, 9.0, 9.1, 9.3, 9.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_parallel, cti_settings=arctic_parallel)
                image_difference_1 = image_post_cti - image_pre_cti

                # Update ci_pre_ctis ci_data so that it corrects the cti added ci_pre_ctis
                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(),
                                        array=image_post_cti)

                image_correct_cti = im.correct_cti_from_image(cti_params=params_parallel, cti_settings=arctic_parallel)
                image_difference_2 = image_correct_cti - image_pre_cti

                assert (abs(image_difference_2) <= abs(image_difference_1)).all()

        class TestSerialCTI:

            def test__array_of_values__corrected_image_more_like_original(self, arctic_serial, params_serial):
                image_pre_cti = np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 3.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 352, 9.4, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 0.0, 9.0, 0.0],
                                          [0.0, 9.0, 9.1, 9.3, 9.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_serial, cti_settings=arctic_serial)
                image_difference_1 = image_post_cti - image_pre_cti

                # Update ci_pre_ctis ci_data so that it corrects the cti added ci_pre_ctis
                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(),
                                        array=image_post_cti)

                image_correct_cti = im.correct_cti_from_image(cti_params=params_serial, cti_settings=arctic_serial)
                image_difference_2 = image_correct_cti - image_pre_cti

                assert (abs(image_difference_2) <= abs(image_difference_1)).all()


class TestQuadrantGeometryEuclidTopRight:
    class TestAddCTI:
        class TestParallelCTI:

            def test__horizontal_charge_line__loses_charge_and_trails_form_above_line(self, arctic_parallel,
                                                                                      params_parallel):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, :] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_parallel, cti_settings=arctic_parallel)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[3:5, :] == 0.0).all()  # No trails closest to read out
                assert (image_difference[2, :] < 0.0).all()  # line loses charge
                assert (image_difference[0:2, :] > 0.0).all()  # Trails appear away from read-out

            def test__vertical_charge_line__loses_charge_no_trails(self, arctic_parallel, params_parallel):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[:, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_parallel, cti_settings=arctic_parallel)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[:, 0:2] == 0.0).all()  # No trails closest to read out
                assert (image_difference[:, 2] < 0.0).all()  # line loses charge
                assert (image_difference[:,
                        3:5] == 0.0).all()  # No trails away from read-out either, as wrong direction

        class TestSerialCTI:

            def test__horizontal_charge_line__loses_charge_no_trails(self, arctic_serial, params_serial):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, :] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_serial, cti_settings=arctic_serial)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[0:2, :] == 0.0).all()  # No trails closest to read out
                assert (image_difference[2, :] < 0.0).all()  # line loses charge
                assert (image_difference[3:5,
                        :] == 0.0).all()  # No trails away from read-out either, as wrong direction

            def test__vertical_charge_line__loses_charge_and_trails_form_to_right_of_line(self, arctic_serial,
                                                                                          params_serial):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[:, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_serial, cti_settings=arctic_serial)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[:, 0:2] > 0.0).all()  # No trails closest to read out
                assert (image_difference[:, 2] < 0.0).all()  # line loses charge
                assert (image_difference[:, 3:5] == 0.0).all()  # Trails appear away from read-out

        class TestParallelAndSerialCTI:

            def test__horizontal_charge_line__loses_charge_trails_form_both_directions(self, arctic_both,
                                                                                       params_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 1:4] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[3:5, :] == 0.0).all()  # No change in front of charge
                assert (image_difference[2, 1:4] < 0.0).all()  # charge lost in charge

                assert (image_difference[0:2, 1:4] > 0.0).all()  # Parallel trails behind charge

                assert (image_difference[0:2, 0] > 0.0).all()  # serial cti trail to left including parallel trails
                assert (image_difference[2:5, 4] == 0.0).all()  # no trails to right

                assert (image_difference[1, 1:4] > image_difference[0, 1:4]).all()  # check parallel cti trails
                # decreasing.

                assert (image_difference[1, 0] > image_difference[
                    0, 0])  # Check serial trails of paralle trails decreasing.

            def test__vertical_charge_line__loses_charge_trails_form_in_serial_direction(self, arctic_both,
                                                                                         params_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[1:4, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[4, 0:5] == 0.0).all()  # No change in front of charge
                assert (image_difference[1:4, 2] < 0.0).all()  # charge lost in charge

                assert (image_difference[0, 2] > 0.0).all()  # Parallel trail behind charge

                assert (image_difference[0:4,
                        0:2] > 0.0).all()  # serial cti trail is to the left including parallel trails
                assert (image_difference[0:5, 3:5] == 0.0).all()  # no serial cti trail to the right

                assert (image_difference[2, 1] > image_difference[2, 0])  # Check serial trails decreasing.
                assert (image_difference[1, 1] > image_difference[
                    1, 0])  # Check serial trails of parallel trails decreasing.

            def test__individual_pixel__trails_form_cross_around_it(self, params_both, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)

                image_difference = image_post_cti - image_pre_cti

                assert (image_difference[3:5, :] == 0.0).all()  # First two rows should all remain zero
                assert (image_difference[:, 3:5] == 0.0).all()  # First tow columns should all remain zero
                assert (image_difference[2, 2] < 0.0)  # pixel which had charge should lose it due to cti.
                assert (image_difference[0:2, 2] > 0.0).all()  # Parallel trail increases charge above pixel
                assert (image_difference[2, 0:2] > 0.0).all()  # Serial trail increases charge to right of pixel
                assert (image_difference[0:2,
                        0:2] > 0.0).all()  # Serial trailing of parallel trail increases charge up-right of pixel

            def test__individual_pixel_double_density__more_captures_so_brighter_trails(self, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                params_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_0 = im.add_cti_to_image(cti_params=params_0, cti_settings=arctic_both)

                params_1 = setup(include_parallel=True, p_trap_densities=(0.2,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.2,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_1 = im.add_cti_to_image(cti_params=params_1, cti_settings=arctic_both)

                image_difference = image_post_cti_1 - image_post_cti_0

                assert (image_difference[3:5, :] == 0.0).all()  # First two rows remain zero
                assert (image_difference[:, 3:5] == 0.0).all()  # First tow columns remain zero
                assert (image_difference[
                            2, 2] < 0.0)  # More captures in second ci_pre_ctis, so more charge in central pixel lost
                assert (image_difference[0:2,
                        2] > 0.0).all()  # More captpures in ci_pre_ctis 2, so brighter parallel trail
                assert (image_difference[2,
                        0:2] > 0.0).all()  # More captpures in ci_pre_ctis 2, so brighter serial trail
                assert (image_difference[0:2, 0:2] > 0.0).all()  # Brighter serial trails from parallel trail trails

            def test__individual_pixel_increase_lifetime_longer_release_so_fainter_trails(self, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                params_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_0 = im.add_cti_to_image(cti_params=params_0, cti_settings=arctic_both)

                params_1 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(20.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(20.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_1 = im.add_cti_to_image(cti_params=params_1, cti_settings=arctic_both)

                image_difference = image_post_cti_1 - image_post_cti_0

                assert (image_difference[3:5, :] == 0.0).all()  # First two rows remain zero
                assert (image_difference[:, 3:5] == 0.0).all()  # First tow columns remain zero
                assert (image_difference[2, 2] == 0.0)  # Same density so same captures
                assert (image_difference[0:2,
                        2] < 0.0).all()  # Longer release in ci_pre_ctis 2, so fainter parallel trail
                assert (image_difference[2,
                        0:2] < 0.0).all()  # Longer release in ci_pre_ctis 2, so fainter serial trail
                assert (image_difference[0:2,
                        0:2] < 0.0).all()  # Longer release in ci_pre_ctis 2, so fainter parallel trail trails

            def test__individual_pixel_increase_beta__fewer_captures_so_fainter_trails(self, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = + 100

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                params_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.8,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.8)

                image_post_cti_0 = im.add_cti_to_image(cti_params=params_0, cti_settings=arctic_both)

                params_1 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.0000001, p_well_fill_beta=0.9,
                                 include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.0000001, s_well_fill_beta=0.9)

                image_post_cti_1 = im.add_cti_to_image(cti_params=params_1, cti_settings=arctic_both)

                image_difference = image_post_cti_1 - image_post_cti_0

                assert (image_difference[3:5, :] == 0.0).all()  # First two rows remain zero
                assert (image_difference[:, 3:5] == 0.0).all()  # First tow columns remain zero
                assert (image_difference[2, 2] > 0.0)  # Higher beta in 2, so fewer captures
                assert (image_difference[0:2, 2] < 0.0).all()  # Fewer catprues in 2, so fainter parallel trail
                assert (image_difference[2, 0:2] < 0.0).all()  # Fewer captures in 2, so fainter serial trail
                assert (image_difference[0:2, 0:2] < 0.0).all()  # fewer captures in 2, so fainter trails trail region

    class TestCorrectCTI:
        class TestParallelCTI:

            def test__array_of_values__corrected_image_more_like_original(self, arctic_parallel, params_parallel):
                image_pre_cti = np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 3.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 352, 9.4, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 0.0, 9.0, 0.0],
                                          [0.0, 9.0, 9.1, 9.3, 9.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_parallel, cti_settings=arctic_parallel)
                image_difference_1 = image_post_cti - image_pre_cti

                # Update ci_pre_ctis ci_data so that it corrects the cti added ci_pre_ctis
                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(),
                                        array=image_post_cti)

                image_correct_cti = im.correct_cti_from_image(cti_params=params_parallel, cti_settings=arctic_parallel)
                image_difference_2 = image_correct_cti - image_pre_cti

                assert (abs(image_difference_2) <= abs(image_difference_1)).all()

        class TestSerialCTI:

            def test__array_of_values__corrected_image_more_like_original(self, arctic_serial, params_serial):
                image_pre_cti = np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 3.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 352, 9.4, 0.0],
                                          [0.0, 9.0, 5.0, 9.5, 0.0, 9.0, 0.0],
                                          [0.0, 9.0, 9.1, 9.3, 9.2, 9.0, 0.0],
                                          [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]])

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_serial, cti_settings=arctic_serial)
                image_difference_1 = image_post_cti - image_pre_cti

                # Update ci_pre_ctis ci_data so that it corrects the cti added ci_pre_ctis
                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(),
                                        array=image_post_cti)

                image_correct_cti = im.correct_cti_from_image(cti_params=params_serial, cti_settings=arctic_serial)
                image_difference_2 = image_correct_cti - image_pre_cti

                assert (abs(image_difference_2) <= abs(image_difference_1)).all()

        class TestSerialAndParallelCTI:

            def test__simple_case__corrected_image_more_like_original(self, params_both, arctic_both):
                image_pre_cti = np.zeros((5, 5))
                image_pre_cti[2, 2] = 1000.

                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(), array=image_pre_cti)

                image_post_cti = im.add_cti_to_image(cti_params=params_both, cti_settings=arctic_both)
                image_difference_1 = image_post_cti - image_pre_cti

                # Update ci_pre_ctis ci_data so that it corrects the cti added ci_pre_ctis
                im = cti_image.ImageFrame(frame_geometry=ci_frame.QuadGeometryEuclid.top_right(),
                                        array=image_post_cti)

                image_correct_cti = im.correct_cti_from_image(cti_params=params_both, cti_settings=arctic_both)
                image_difference_2 = image_correct_cti - image_pre_cti

                assert (abs(image_difference_2) <= abs(image_difference_1)).all()
